# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

import chartify
import numpy as np
import pandas as pd
from scipy.stats import beta

from spotify_confidence.analysis.bayesian.bayesian_base import (
    BaseTest,
    randomization_warning_decorator,
    format_str_precision,
    axis_format_precision,
)

from spotify_confidence.analysis.confidence_utils import de_list_if_length_one


class BetaBinomial(BaseTest):
    def __init__(
        self,
        data_frame,
        numerator_column,
        denominator_column,
        categorical_group_columns=None,
        ordinal_group_column=None,
        prior_alpha_column=None,
        prior_beta_column=None,
        interval_size=0.95,
    ):
        """
        Bayesian BetaBinomial model.

        See: https://en.wikipedia.org/wiki/Beta-binomial_distribution

        data_frame (pd.DataFrame): DataFrame
        numerator_column (str): Column name for numerator column.
        denominator_column (str): Column name for denominator column.
        categorical_group_columns (str or list): Column names
            for categorical groupings.
        ordinal_group_column (str): Column name for ordinal
            grouping (e.g. numeric or date values).
        prior_alpha_column (str): Column name to use for prior alpha.
        prior_beta_column (str): Column name to use for prior beta.
        interval_size (float): Size of credible intervals. Default 0.95
        """
        super().__init__(
            data_frame,
            categorical_group_columns,
            ordinal_group_column,
            numerator_column,
            denominator_column,
            interval_size,
        )

        self._monte_carlo_sample_size = 500000

        # Initialize priors.
        if prior_alpha_column is None or prior_beta_column is None:
            self._alpha_prior, self._beta_prior = (0.5, 0.5)
        else:
            self._alpha_prior = data_frame[prior_alpha_column]
            self._beta_prior = data_frame[prior_beta_column]

    def _interval(self, row):
        interval = beta.interval(
            self._interval_size,
            row[self._numerator_column] + self._alpha_prior,
            row[self._denominator_column] - row[self._numerator_column] + self._beta_prior,
        )
        return interval

    def _posterior_parameters(self, group_df):
        """Calculate parameters of posterior distribution.

        Returns:
            tuple of floats: posterior_alpha, posterior_beta"""
        numerator = group_df[self._numerator_column].values[0]
        denominator = group_df[self._denominator_column].values[0]
        posterior_alpha = numerator + self._alpha_prior
        posterior_beta = denominator - numerator + self._beta_prior
        return posterior_alpha, posterior_beta

    def _beta_pdf(self, group_df):
        """Beta pdfs for the given dataframe"""
        posterior_alpha, posterior_beta = self._posterior_parameters(group_df)
        epsilon = 0.001
        lower_range = beta.isf(1.0 - epsilon, posterior_alpha, posterior_beta)
        upper_range = beta.isf(epsilon, posterior_alpha, posterior_beta)
        x_range = np.linspace(lower_range, upper_range, 1000)
        beta_pdf = [beta.pdf(x, posterior_alpha, posterior_beta) for x in x_range]
        beta_dist = pd.DataFrame({"x": x_range, "y": beta_pdf})
        return beta_dist

    def _sample_posterior(self, group_df, posterior_sample_size=None):
        """MCMC sampling of posterior distribution.
        Used to calculate the posterior distribution of
        the difference in Beta RVs.

        Arguments:
            - seed (int): Seed for random number generator.
                Set it to make the posteriors deterministic.
            - posterior_sample_size (int): Number of posterior
                samples (affects precision)
        """
        if posterior_sample_size is None:
            posterior_sample_size = self._monte_carlo_sample_size
        posterior_alpha, posterior_beta = self._posterior_parameters(group_df)
        posterior_samples = np.random.beta(posterior_alpha, posterior_beta, size=posterior_sample_size)
        return posterior_samples

    def _categorical_summary_plot(self, level_name, level_df, remaining_groups, groupby):
        if not remaining_groups:
            remaining_groups = groupby

        grouped_df = level_df.groupby(de_list_if_length_one(remaining_groups))

        distributions = pd.DataFrame()
        for group_name, group_df in grouped_df:
            beta_dist = self._beta_pdf(group_df)
            beta_dist["group"] = str(group_name)
            distributions = pd.concat([distributions, beta_dist], axis=0)

        # Filter out the long tails of the distributions
        filtered_xs = distributions.groupby("x")["y"].max().reset_index().loc[lambda x: x["y"] > 0.01]
        distributions = distributions[distributions["x"].isin(filtered_xs["x"])]

        # Remove legend if only one color
        color_column = "group" if len(grouped_df) > 1 else None

        ch = chartify.Chart()
        ch.plot.area(
            distributions,
            "x",
            "y",
            color_column=color_column,
            stacked=False,
            color_order=[str(x) for x in list(grouped_df.groups.keys())],
        )
        ch.set_title("Estimate of {} / {}".format(self._numerator_column, self._denominator_column))

        if groupby:
            ch.set_subtitle("{}: {}".format(groupby, level_name))
        else:
            ch.set_subtitle("")
        ch.axes.set_xaxis_label("{} / {}".format(self._numerator_column, self._denominator_column))
        ch.axes.set_yaxis_label("Probability Density")
        ch.set_source_label("")
        ch.axes.set_yaxis_range(0)
        axis_format = axis_format_precision(distributions["x"].min(), distributions["x"].max(), absolute=True)
        ch.axes.set_xaxis_tick_format(axis_format)

        ch.style.color_palette.reset_palette_order()

        # Plot callouts for the means
        for group_name, group_df in grouped_df:
            posterior_alpha, posterior_beta = self._posterior_parameters(group_df)
            posterior_mean = posterior_alpha / (posterior_alpha + posterior_beta)
            density = beta.pdf(posterior_mean, posterior_alpha, posterior_beta)
            ch.callout.line(
                posterior_mean,
                orientation="height",
                line_color=ch.style.color_palette.next_color(),
                line_dash="dashed",
            )
            ch.callout.text(
                f"{posterior_mean:{format_str_precision(posterior_mean, absolute=False)}}", posterior_mean, density
            )

        ch.axes.hide_yaxis()
        if color_column:
            ch.set_legend_location("outside_bottom")
        return ch

    def _difference_posteriors(self, data, level_1, level_2, absolute=True):
        posterior_1 = self._sample_posterior(data.get_group(level_1))
        posterior_2 = self._sample_posterior(data.get_group(level_2))

        if absolute:
            difference_posterior = posterior_2 - posterior_1
        else:
            difference_posterior = posterior_2 / posterior_1 - 1.0

        return difference_posterior

    def _differences(self, difference_posterior, level_1, level_2, absolute):
        # 95% credible interval for posterior
        credible_interval = (
            pd.Series(difference_posterior).quantile((1.0 - self._interval_size) / 2),
            pd.Series(difference_posterior).quantile((1.0 - self._interval_size) / 2 + self._interval_size),
        )

        # Probability that posterior is greater
        # than zero (count occurences in the MC sample)
        p_gt_zero = (difference_posterior > 0).mean()

        expected_loss_v2 = difference_posterior[difference_posterior < 0].sum() / len(difference_posterior)
        if (difference_posterior > 0).sum() == 0:
            expected_gain_v2 = 0
        else:
            expected_gain_v2 = difference_posterior[difference_posterior > 0].sum() / len(difference_posterior)

        expected_loss_v1 = (difference_posterior[difference_posterior * -1.0 < 0] * -1.0).sum() / len(
            difference_posterior
        )

        if (difference_posterior * -1.0 > 0).sum() == 0:
            expected_gain_v1 = 0
        else:
            expected_gain_v1 = (difference_posterior[difference_posterior * -1.0 > 0] * -1.0).sum() / len(
                difference_posterior
            )

        return pd.DataFrame(
            OrderedDict(
                [
                    ("level_1", str(level_1)),
                    ("level_2", str(level_2)),
                    ("absolute_difference", absolute),
                    ("difference", difference_posterior.mean()),
                    ("ci_lower", [credible_interval[0]]),
                    ("ci_upper", [credible_interval[1]]),
                    ("P(level_2 > level_1)", p_gt_zero),
                    ("level_1 potential loss", expected_loss_v1),
                    ("level_1 potential gain", expected_gain_v1),
                    ("level_2 potential loss", expected_loss_v2),
                    ("level_2 potential gain", expected_gain_v2),
                ]
            )
        )

    def _difference(self, level_name, level_df, remaining_groups, groupby, level_1, level_2, absolute):
        difference_df, _ = self._difference_and_difference_posterior(
            level_df, remaining_groups, level_2, level_1, absolute
        )

        self._add_group_by_columns(difference_df, groupby, level_name)

        return difference_df

    def _difference_and_difference_posterior(self, level_df, remaining_groups, level_2, level_1, absolute):
        self._validate_levels(level_df, remaining_groups, level_1)
        self._validate_levels(level_df, remaining_groups, level_2)
        # difference is posterior_2 - posterior_1
        difference_posterior = self._difference_posteriors(
            level_df.groupby(remaining_groups), level_1, level_2, absolute
        )
        difference_df = self._differences(difference_posterior, level_1, level_2, absolute)
        return difference_df, difference_posterior

    @randomization_warning_decorator
    def difference(self, level_1, level_2, absolute=True, groupby=None):
        """Return DataFrame with summary statistics of the difference between
        level 1 and level 2.

        Args:
            level_1 (str, tuple of str): Name of first level.
            level_2 (str, tuple of str): Name of second level.
            absolute (bool): If True then return the
                absolute difference (level2 - level1)
                otherwise return the relative difference (level2 / level1 - 1)
            groupby (str): Name of column.
                If specified, will return the difference for each level
                of the grouped dimension.

        Returns:
            Pandas DataFrame with the following columns:
            - level_1: Name of level 1.
            - level_2: Name of level 2.
            - absolute_difference: True if absolute.
                Absolute: level2 - level1
                Relative: level2 / level1 - 1
            - difference: Best estimate of the difference between level 2 and 1.
                Posterior mean of the difference between level 1 and level 2.
                https://en.wikipedia.org/wiki/Bayes_estimator
            - ci_lower: Lower credible interval bound of the difference.
            - ci_upper: Upper credible interval bound of the difference.
            - P(level_2 > level_1): Probability that the level 2 > level 1.
            - level_1 potential loss: The expected loss if we
                switch to level 1, but level 2 is actually better.
            - level_1 potential gain: The expected gain if we
                switch to level 1, and level 1 is actually better.
            - level_2 potential loss: The expected loss if we
                switch to level 2, but level 1 is actually better.
            - level_2 potential gain: The expected gain if we
                switch to level 2, and level 2 is actually better.
        """

        results_df = self._iterate_groupby_to_dataframe(
            self._difference, groupby=groupby, level_1=level_1, level_2=level_2, absolute=absolute
        )

        return results_df

    @randomization_warning_decorator
    def _categorical_difference_plot(self, level_1, level_2, absolute, groupby):
        chart_grid = self._iterate_groupby_to_chartgrid(
            self._categorical_difference_plot_, groupby=groupby, level_1=level_1, level_2=level_2, absolute=absolute
        )

        return chart_grid

    def _categorical_difference_plot_(
        self, level_name, level_df, remaining_groups, groupby, level_1, level_2, absolute
    ):
        difference_df, difference_posterior = self._difference_and_difference_posterior(
            level_df, remaining_groups, level_2, level_1, absolute
        )

        posterior_mean = difference_df["difference"][0]
        # potential_loss = difference_df['{} potential loss'.format(level_2)][0]

        # Take the difference posterior and create a chart
        df = pd.DataFrame({"values": difference_posterior})

        ch = chartify.Chart(y_axis_type="density", x_axis_type="linear")

        ch.plot.kde(df, "values")

        ch.set_title("Change from {} to {}".format(level_1, level_2))

        subtitle = "" if not groupby else "{}: {}".format(groupby, level_name)
        ch.set_subtitle(subtitle)

        # Line at no difference
        ch.callout.line(0, orientation="height", line_color="black", line_dash="dashed")
        # ch.callout.text('No change', 0, .5, angle=90)

        # Plot callout for the mean
        ch.callout.line(
            posterior_mean, orientation="height", line_color=ch.style.color_palette._colors[0], line_dash="dashed"
        )
        # ch.callout.text(
        #     '{0:.2f}%'.format(posterior_mean * 100), posterior_mean, 0)
        ch.callout.text(
            f"Expected change: {posterior_mean:{format_str_precision(posterior_mean, absolute=False)}}",
            posterior_mean,
            0,
            angle=90,
        )

        # ch.callout.line(
        #     potential_loss,
        #     orientation='height',
        #     line_color=ch.style.color_palette._colors[1])
        # ch.callout.text(
        #     'Potential Loss: {0:.2f}%'.format(potential_loss * 100),
        #     potential_loss,
        #     1.5,
        #     angle=90)
        # ch.callout.text(
        #     '{0:.2f}%'.format(potential_loss * 100), potential_loss, 1.)

        ch.set_source_label("")
        ch.axes.set_yaxis_range(0)
        ch.axes.set_xaxis_label(self.get_difference_plot_label(absolute))
        ch.axes.set_yaxis_label("Probability Density")
        ch.axes.hide_yaxis()
        axis_format = axis_format_precision(df["values"].max() * 10, df["values"].min() * 10, absolute)
        ch.axes.set_xaxis_tick_format(axis_format)

        return ch

    def _multiple_difference_joint_dataframe(self, *args, **kwargs):
        return self._multiple_difference_joint_base(*args, **kwargs)[0]

    def _multiple_difference_joint_base(self, level_name, level_df, remaining_groups, groupby, level, absolute):
        grouped_df = level_df.groupby(remaining_groups)

        grouped_df_keys = tuple(grouped_df.groups.keys())

        self._validate_levels(level_df, remaining_groups, level)

        posteriors = [self._sample_posterior(grouped_df.get_group(level)) for level in grouped_df_keys]

        var_indx = grouped_df_keys.index(level)
        other_indx = [i for i, value in enumerate(grouped_df_keys) if value != level]

        posterior_matrix = np.vstack(posteriors)

        ge_bool_matrix = posterior_matrix[var_indx, :] >= posterior_matrix[:, :]

        best_arr = ge_bool_matrix.all(axis=0)

        p_ge_all = best_arr.mean()

        end_value = posterior_matrix[var_indx]
        start_value = posterior_matrix[other_indx].max(axis=0)

        if absolute:
            difference_posterior = end_value - start_value
        else:
            difference_posterior = end_value / start_value - 1

        # E(level - best level | level != best)
        if not (~best_arr).sum():
            expected_loss = 0
        else:
            expected_loss = difference_posterior[~best_arr].mean()

        # E(level - median level | level = best)
        if not (best_arr).sum():
            expected_gain = 0
        else:
            expected_gain = difference_posterior[best_arr].mean()

        expectation = difference_posterior.mean()
        ci_l_expectation = pd.Series(difference_posterior).quantile((1.0 - self._interval_size) / 2)
        ci_u_expectation = pd.Series(difference_posterior).quantile(
            (1.0 - self._interval_size) / 2 + self._interval_size
        )

        difference_df = pd.DataFrame(
            OrderedDict(
                [
                    ("level", [str(level)]),
                    ("absolute_difference", absolute),
                    ("difference", expectation),
                    ("ci_lower", ci_l_expectation),
                    ("ci_upper", ci_u_expectation),
                    ("P({} >= all)".format(level), p_ge_all),
                    ("{} potential loss".format(level), expected_loss),
                    ("{} potential gain".format(level), expected_gain),
                ]
            )
        )
        self._add_group_by_columns(difference_df, groupby, level_name)

        return (difference_df, difference_posterior)

    @randomization_warning_decorator
    def multiple_difference_joint(self, level, absolute=True, groupby=None):
        """Calculate the joint probability that the given level is greater
        than all other levels in the test.

        Args:
            level (str, tuple of str): Name of level.
            absolute (bool): If True then return the absolute difference
                otherwise return the relative difference.
            groupby (str): Name of column.
                If specified, will return an interval for each level
                of the grouped dimension.

        Returns:
            Pandas DataFrame with the following columns:
            - level: Name of level
            - absolute_difference: True if absolute.
                Absolute: level2 - level1
                Relative: level2 / level1 - 1
            - difference: Difference between the level and the best performing
                among the other levels.
            - ci_lower: Lower credible interval bound of the difference.
            - ci_upper: Upper credible interval bound of the difference.
            - P(level > all): Probability that the level > all other levels.
            - potential loss: The expected loss if we
                switch to level, but some other level is actually better.
            - potential gain: The expected gain if we
                switch to level, and it is actually the best.
        """

        results_df = self._iterate_groupby_to_dataframe(
            self._multiple_difference_joint_dataframe, groupby=groupby, level=level, absolute=absolute
        )

        return results_df

    def _multiple_difference_joint_plot(self, level_name, level_df, remaining_groups, groupby, level, absolute):
        self._validate_levels(level_df, remaining_groups, level)

        difference_df, difference_posterior = self._multiple_difference_joint_base(
            level_name, level_df, remaining_groups, groupby, level, absolute
        )

        posterior_mean = difference_df.loc[:, "difference"].values[0]

        # potential_loss = difference_df.loc[:, '{} potential loss'.format(
        #     level)].values[0]

        # Take the difference posterior and create a chart
        df = pd.DataFrame({"values": difference_posterior})

        ch = chartify.Chart(y_axis_type="density", x_axis_type="linear")

        ch.plot.kde(df, "values")

        ch.set_title("Comparison to {}".format(level))

        subtitle = "" if not groupby else "{}: {}".format(groupby, level_name)
        ch.set_subtitle(subtitle)

        # Line at no difference
        ch.callout.line(0, orientation="height", line_color="black")

        # Plot callout for the mean
        ch.callout.line(posterior_mean, orientation="height", line_color=ch.style.color_palette._colors[0])

        ch.callout.text(
            f"Expected change: {posterior_mean:{format_str_precision(posterior_mean, absolute=False)}}",
            posterior_mean,
            0,
            angle=90,
        )

        ch.set_source_label("")
        ch.axes.set_yaxis_range(0)
        ch.axes.set_xaxis_label(self.get_difference_plot_label(absolute))
        ch.axes.set_yaxis_label("Probability Density")
        ch.axes.hide_yaxis()

        axis_format = axis_format_precision(df["values"].max() * 10, df["values"].min() * 10, absolute)
        ch.axes.set_xaxis_tick_format(axis_format)
        return ch

    @randomization_warning_decorator
    def multiple_difference_joint_plot(self, level, absolute=True, groupby=None):
        """Calculate the joint probability that the given level is greater
        than all other levels in the test.

        Args:
            level (str, tuple of str): Name of level.
            absolute (bool): If True then return the absolute difference
                otherwise return the relative difference.
            groupby (str): Name of column.
                If specified, will return an interval for each level
                of the grouped dimension.

        Returns:
            GroupedChart object.
        """

        results_df = self._iterate_groupby_to_chartgrid(
            self._multiple_difference_joint_plot, groupby=groupby, level=level, absolute=absolute
        )

        return results_df

    def _multiple_difference(
        self, level_name, level_df, remaining_groups, groupby, level, absolute, level_as_reference
    ):
        grouped_df = level_df.groupby(remaining_groups)

        grouped_df_keys = tuple(grouped_df.groups.keys())

        other_keys = [value for i, value in enumerate(grouped_df_keys) if value != level]

        for key in other_keys:
            # Switch the subtraction order as specified.
            start_value, end_value = level, key
            if not level_as_reference:
                start_value, end_value = end_value, start_value

            difference_df = self._difference(
                level_name, level_df, remaining_groups, groupby, start_value, end_value, absolute=absolute
            )

            yield difference_df

    @randomization_warning_decorator
    def multiple_difference(self, level, absolute=True, groupby=None, level_as_reference=False):
        """Pairwise comparison of the given level to all others.

        Args:
            level (str, tuple of str): Name of level.
            absolute (bool): If True then return the absolute difference
                otherwise return the relative difference.
            groupby (str): Name of column.
                If specified, will return an interval for each level
                of the grouped dimension.
            level_as_reference (bool): If True, the given level is the reference
                value for the change. (level1)

        Returns:
            Pandas DataFrame with the following columns:
            - groupby (If groupby is not None): Grouped dimension
            - level_1: Name of level 1.
            - level_2: Name of level 2.
            - absolute_difference: True if absolute.
                Absolute: level2 - level1
                Relative: level2 / level1 - 1
            - difference: Best estimate of the difference between level 2 and 1.
                Posterior mean of the difference between level 1 and level 2.
                https://en.wikipedia.org/wiki/Bayes_estimator
            - ci_lower: Lower credible interval bound of the difference.
            - ci_upper: Upper credible interval bound of the difference.
            - P(level_2 > level_1): Probability that the level 2 > level 1.
            - level_1 potential loss: The expected loss if we
                switch to level 1, but level 2 is actually better.
            - level_1 potential gain: The expected gain if we
                switch to level 1, and level 1 is actually better.
            - level_2 potential loss: The expected loss if we
                switch to level 2, but level 1 is actually better.
            - level_2 potential gain: The expected gain if we
                switch to level 2, and level 2 is actually better.
        """

        results_df = self._iterate_groupby_to_dataframe(
            self._multiple_difference,
            groupby=groupby,
            level=level,
            absolute=absolute,
            level_as_reference=level_as_reference,
        )

        results_df = results_df.reset_index(drop=True)

        return results_df

    def _categorical_multiple_difference_chart(
        self, level_name, level_df, remaining_groups, groupby, level, absolute, level_as_reference
    ):
        grouped_df = level_df.groupby(remaining_groups)

        grouped_df_keys = tuple(grouped_df.groups.keys())

        self._validate_levels(level_df, remaining_groups, level)

        posteriors = [self._sample_posterior(grouped_df.get_group(level)) for level in grouped_df_keys]

        var_indx = grouped_df_keys.index(level)

        other_indx = [i for i, value in enumerate(grouped_df_keys) if value != level]

        posterior_matrix = np.vstack(posteriors)

        start_value = posterior_matrix[var_indx]
        end_value = posterior_matrix
        if not level_as_reference:
            start_value, end_value = end_value, start_value

        if absolute:
            difference_posterior = end_value - start_value
        else:
            difference_posterior = end_value / start_value - 1

        df = pd.DataFrame()
        for group in other_indx:
            df = pd.concat(
                [df, pd.DataFrame({"values": difference_posterior[group], "group": str(grouped_df_keys[group])})],
                axis=0,
            )

        # Take the difference posterior and create a chart
        # df = pd.DataFrame({'values': difference_posterior})

        ch = chartify.Chart(y_axis_type="density", x_axis_type="linear")

        ch.plot.kde(df, "values", color_column="group")

        title_change_label = "from" if level_as_reference else "to"
        ch.set_title("Change {} {}".format(title_change_label, level))

        subtitle = "" if not groupby else "{}: {}".format(groupby, level_name)
        ch.set_subtitle(subtitle)

        # Line at no difference
        ch.callout.line(0, orientation="height", line_color="black", line_dash="dashed")
        # ch.callout.text('No change', 0, .5, angle=90)
        ch.style.color_palette.reset_palette_order()

        for group in other_indx:
            posterior_mean = difference_posterior[group].mean()
            # Plot callout for the mean
            ch.callout.line(
                posterior_mean,
                orientation="height",
                line_color=ch.style.color_palette.next_color(),
                line_dash="dashed",
            )

            ch.callout.text(
                f"Expected change: {posterior_mean:{format_str_precision(posterior_mean, absolute=False)}}",
                posterior_mean,
                0,
                angle=90,
            )

        # ch.callout.line(
        #     potential_loss,
        #     orientation='height',
        #     line_color=ch.style.color_palette._colors[1])
        # ch.callout.text(
        #     'Potential Loss: {0:.2f}%'.format(potential_loss * 100),
        #     potential_loss,
        #     1.5,
        #     angle=90)
        # ch.callout.text(
        #     '{0:.2f}%'.format(potential_loss * 100), potential_loss, 1.)

        ch.set_source_label("")
        ch.axes.set_yaxis_range(0)
        ch.axes.set_xaxis_label(self.get_difference_plot_label(absolute))
        ch.axes.set_yaxis_label("Probability Density")
        ch.axes.hide_yaxis()
        axis_format = axis_format_precision(df["values"].max() * 10, df["values"].min() * 10, absolute)
        ch.axes.set_xaxis_tick_format(axis_format)

        return ch

    @randomization_warning_decorator
    def _categorical_multiple_difference_plot(self, level, absolute, groupby, level_as_reference):
        """Pairwise comparison of the given level to all others.

        Args:
            level (str, tuple of str): Name of level.
            absolute (bool): If True then return the absolute difference
                otherwise return the relative difference.
            groupby (str): Name of column.
                If specified, will return an interval for each level
                of the grouped dimension.
            level_as_reference (bool): If True, the given level is the reference
                value for the change. (level1)

        Returns:
            GroupedChart object.
        """

        results_df = self._iterate_groupby_to_chartgrid(
            self._categorical_multiple_difference_chart,
            groupby=groupby,
            level=level,
            absolute=absolute,
            level_as_reference=level_as_reference,
        )

        return results_df


# class GammaPoisson(PoissonResponse):
#     pass


# class DirichetMultinomial(MultinomialResponse):
#     def __init__(self,
#                  data_frame,
#                  group_columns,
#                  category_column,
#                  value_column,
#                  prior_value_column=None):

#         super().__init__(data_frame, group_columns, category_column,
#                          value_column)


# class Gaussian(GaussianResponse):
#     def __init__(self,
#                  data_frame,
#                  groupings,
#                  mean_col,
#                  std_col,
#                  n_col,
#                  time_grouping=None,
#                  prior_columns=None):
#         self.prior_lambda_column = prior_lambda_column
#         super(BaseGaussianResponse, self).__init__(
#             data_frame, groups, mean_col, std_col, n_col, time_grouping)
#         raise (NotImplementedError)


# class DirichetCategorical(CategoricalResponse):
#     pass



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pkg_resources import require as _require
from .analysis.bayesian.bayesian_models import BetaBinomial
from spotify_confidence.analysis.frequentist.chi_squared import ChiSquared
from spotify_confidence.analysis.frequentist.t_test import StudentsTTest
from spotify_confidence.analysis.frequentist.z_test import ZTest
from spotify_confidence.analysis.frequentist.z_test_linreg import ZTestLinreg
from spotify_confidence.analysis.frequentist.experiment import Experiment
from spotify_confidence.analysis.frequentist.sample_size_calculator import SampleSizeCalculator
from .samplesize.sample_size_calculator import SampleSize

from . import examples
from .options import options

__version__ = _require("spotify_confidence")[0].version

__all__ = [
    "BetaBinomial",
    "ChiSquared",
    "StudentsTTest",
    "ZTest",
    "ZTestLinreg",
    "Experiment",
    "SampleSizeCalculator",
    "examples",
    "options",
    "SampleSize",
]



from spotify_confidence.analysis.constants import CHI2, TTEST, ZTEST, BOOTSTRAP, ZTESTLINREG
from spotify_confidence.analysis.frequentist.confidence_computers import (
    chi_squared_computer as chi_squared_computer,
    t_test_computer as t_test_computer,
    z_test_computer as z_test_computers,
    bootstrap_computer as bootstrap_computer,
    z_test_linreg_computer as z_test_linreg_computer,
)

confidence_computers = {
    CHI2: chi_squared_computer,
    TTEST: t_test_computer,
    ZTEST: z_test_computers,
    BOOTSTRAP: bootstrap_computer,
    ZTESTLINREG: z_test_linreg_computer,
}



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Union, Iterable, Tuple, List

import numpy as np
from pandas import DataFrame, concat, Series
from scipy.stats import norm

from spotify_confidence.analysis.constants import (
    SFX1,
    SFX2,
)


def groupbyApplyParallel(dfGrouped, func_to_apply):
    with ThreadPoolExecutor(max_workers=32, thread_name_prefix="groupbyApplyParallel") as p:
        ret_list = p.map(
            func_to_apply,
            [group for name, group in dfGrouped],
        )
    return concat(ret_list)


def applyParallel(df, func_to_apply, splits=32):
    with ThreadPoolExecutor(max_workers=splits, thread_name_prefix="applyParallel") as p:
        ret_list = p.map(
            func_to_apply,
            np.array_split(df, min(splits, len(df))),
        )
    return concat(ret_list)


def get_all_group_columns(categorical_columns: Iterable, additional_column: str) -> Iterable:
    all_columns = listify(categorical_columns) + listify(additional_column)
    return list(OrderedDict.fromkeys(all_columns))


def remove_group_columns(categorical_columns: Iterable, additional_column: str) -> Iterable:
    od = OrderedDict.fromkeys(categorical_columns)
    if additional_column is not None:
        del od[additional_column]
    return list(od)


def validate_categorical_columns(categorical_group_columns: Union[str, Iterable]) -> Iterable:
    if isinstance(categorical_group_columns, str):
        pass
    elif isinstance(categorical_group_columns, Iterable):
        pass
    else:
        raise TypeError(
            """categorical_group_columns must be string or
                           iterable (list of columns) and you must
                           provide at least one"""
        )


def listify(column_s: Union[str, Iterable]) -> List:
    if isinstance(column_s, str):
        return [column_s]
    elif isinstance(column_s, Iterable):
        return list(column_s)
    elif column_s is None:
        return []


def get_remaning_groups(all_groups: Iterable, some_groups: Iterable) -> Iterable:
    if some_groups is None:
        remaining_groups = all_groups
    else:
        remaining_groups = [group for group in all_groups if group not in some_groups and group is not None]
    return remaining_groups


def get_all_categorical_group_columns(
    categorical_columns: Union[str, Iterable, None],
    metric_column: Union[str, None],
    treatment_column: Union[str, None],
) -> Iterable:
    all_columns = listify(treatment_column) + listify(categorical_columns) + listify(metric_column)
    return list(OrderedDict.fromkeys(all_columns))


def validate_levels(df: DataFrame, level_columns: Union[str, Iterable], levels: Iterable):
    for level in levels:
        try:
            df.groupby(level_columns).get_group(level)
        except (KeyError, ValueError):
            raise ValueError(
                """
                    Invalid level: '{}'
                    Must supply a level within the ungrouped dimensions: {}
                    Valid levels:
                    {}
                    """.format(
                    level, level_columns, list(df.groupby(level_columns).groups.keys())
                )
            )


def validate_and_rename_columns(df: DataFrame, columns: Iterable[str]) -> DataFrame:
    for column in columns:
        if column is None or column + SFX1 not in df.columns or column + SFX2 not in df.columns:
            continue

        if (df[column + SFX1].isna() == df[column + SFX2].isna()).all() and (
            df[column + SFX1][df[column + SFX1].notna()] == df[column + SFX2][df[column + SFX2].notna()]
        ).all():
            df = df.rename(columns={column + SFX1: column}).drop(columns=[column + SFX2])
        else:
            raise ValueError(f"Values of {column} do not agree across levels: {df[[column + SFX1, column + SFX2]]}")
    return df


def drop_and_rename_columns(df: DataFrame, columns: Iterable[str]) -> DataFrame:
    columns_dict = {col + SFX1: col for col in columns}
    return df.rename(columns=columns_dict).drop(columns=[col + SFX2 for col in columns])


def level2str(level: Union[str, Tuple]) -> str:
    if isinstance(level, str) or not isinstance(level, Iterable):
        return str(level)
    else:
        return ", ".join([str(sub_level) for sub_level in level])


def validate_data(df: DataFrame, columns_that_must_exist, group_columns: Iterable, ordinal_group_column: str):
    """Integrity check input dataframe."""
    for col in columns_that_must_exist:
        _validate_column(df, col)

    if not group_columns:
        raise ValueError(
            """At least one of `categorical_group_columns`
                            or `ordinal_group_column` must be specified."""
        )

    for col in group_columns:
        _validate_column(df, col)

    # Ensure there's at most 1 observation per grouping.
    max_one_row_per_grouping = all(df.groupby(group_columns, sort=False).size() <= 1)
    if not max_one_row_per_grouping:
        raise ValueError("""Each grouping should have at most 1 observation.""")

    if ordinal_group_column:
        ordinal_column_type = df[ordinal_group_column].dtype.type
        if not np.issubdtype(ordinal_column_type, np.number) and not issubclass(ordinal_column_type, np.datetime64):
            raise TypeError(
                """`ordinal_group_column` is type `{}`.
        Must be number or datetime type.""".format(
                    ordinal_column_type
                )
            )


def _validate_column(df: DataFrame, col: str):
    if col not in df.columns:
        raise ValueError(f"""Column {col} is not in dataframe""")


def is_non_inferiority(nim) -> bool:
    if isinstance(nim, float):
        return not np.isnan(nim)
    elif nim is None:
        return nim is not None


def reset_named_indices(df):
    named_indices = [name for name in df.index.names if name is not None]
    if len(named_indices) > 0:
        return df.reset_index(named_indices, drop=True).sort_index()
    else:
        return df


def _get_finite_bounds(numbers: Series) -> Tuple[float, float]:
    finite_numbers = numbers[numbers.abs() != float("inf")]
    return finite_numbers.min(), finite_numbers.max()


def axis_format_precision(numbers: Series, absolute: bool, extra_zeros: int = 0) -> Tuple[str, float, float]:
    min_value, max_value = _get_finite_bounds(numbers)

    if max_value == min_value:
        return "0.00", min_value, max_value

    extra_zeros += 2 if absolute else 0
    precision = -int(np.log10(abs(max_value - min_value))) + extra_zeros
    zeros = "".join(["0"] * precision)
    return "0.{}{}".format(zeros, "" if absolute else "%"), min_value, max_value


def to_finite(s: Series, lower_limit: float, upper_limit: float) -> Series:
    return s.clip(-100 * abs(lower_limit), 100 * abs(upper_limit))


def add_color_column(df: DataFrame, cols: Iterable) -> DataFrame:
    return df.assign(color=df[cols].agg(level2str, axis="columns"))


def power_calculation(mde: float, baseline_var: float, alpha: float, n1: int, n2: int) -> float:
    z_alpha = norm.ppf(1 - alpha / 2)
    a = abs(mde) / np.sqrt(baseline_var)
    b = np.sqrt(n1 * n2 / (n1 + n2))
    z_stat = a * b

    return norm.cdf(z_stat - z_alpha) + norm.cdf(-z_stat - z_alpha)


def unlist(x):
    x0 = x[0] if isinstance(x, list) else x
    x1 = np.atleast_2d(x0)
    if x1.shape[0] < x1.shape[1]:
        x1 = x1.transpose()
    return x1


def dfmatmul(x, y, outer=True):
    x = np.atleast_2d(x)
    y = np.atleast_2d(y)
    if x.shape[0] < x.shape[1]:
        x = x.transpose()
    if y.shape[0] < y.shape[1]:
        y = y.transpose()

    if outer:
        out = np.matmul(x, np.transpose(y))
    else:
        out = np.matmul(np.transpose(x), y)

    if out.size == 1:
        out = out.item()
    return out


def de_list_if_length_one(x):
    """Return first element of x if x is a list of length one"""
    is_iterable = type(x) != str and isinstance(x, Iterable)
    return x[0] if is_iterable and len(x) == 1 else x



from functools import reduce
from typing import Union, Dict

import numpy as np
from pandas import DataFrame, Series

from spotify_confidence.analysis.confidence_utils import unlist, dfmatmul
from spotify_confidence.analysis.constants import (
    REGRESSION_PARAM,
    FEATURE,
    FEATURE_SUMSQ,
    FEATURE_CROSS,
    NUMERATOR,
    DENOMINATOR,
)
from spotify_confidence.analysis.frequentist.confidence_computers import z_test_computer


def estimate_slope(df, **kwargs: Dict) -> DataFrame:
    if kwargs[FEATURE] not in df:
        return df

    def col_sum(x):
        return reduce(lambda x, y: x + y, x)

    def dimension(x):
        return x.shape[0] if isinstance(x, np.ndarray) and x.size > 1 else 1

    k = df[kwargs[FEATURE_SUMSQ]].apply(dimension).iloc[0]

    XX0 = np.zeros((k + 1, k + 1))
    XX0[1 : (k + 1), 1 : (k + 1)] = col_sum(df[kwargs[FEATURE_SUMSQ]])
    XX0[0, 0] = col_sum(df[kwargs[DENOMINATOR]])
    XX0[0, 1 : (k + 1)] = col_sum(df[kwargs[FEATURE]])
    XX0[1 : (k + 1), 0] = col_sum(df[kwargs[FEATURE]])

    Xy0 = np.zeros((k + 1, 1))
    Xy0[0,] = col_sum(df[kwargs[NUMERATOR]])
    Xy0[1 : (k + 1),] = np.atleast_2d(col_sum(df[kwargs[FEATURE_CROSS]])).reshape(-1, 1)

    b = np.matmul(np.linalg.inv(XX0), Xy0)
    out = b[1 : (k + 1)]
    if out.size == 1:
        out = out.item()

    outseries = Series(index=df.index, dtype=df[kwargs[FEATURE]].dtype)
    df[REGRESSION_PARAM] = outseries.apply(lambda x: out)
    return df


def point_estimate(df: Series, **kwargs) -> float:
    df = estimate_slope(df, **kwargs)
    point_estimate = df[kwargs[NUMERATOR]] / df[kwargs[DENOMINATOR]]

    if REGRESSION_PARAM in df:
        feature_mean = df[kwargs[FEATURE]].sum() / df[kwargs[DENOMINATOR]].sum()

        def lin_reg_point_estimate_delta(row: Series, feature_mean: float, **kwargs: Dict) -> Series:
            return dfmatmul(
                row[REGRESSION_PARAM], row[kwargs[FEATURE]] - feature_mean * row[kwargs[DENOMINATOR]], outer=False
            )

        return (
            point_estimate
            - df.apply(lin_reg_point_estimate_delta, feature_mean=feature_mean, axis=1, **kwargs)
            / df[kwargs[DENOMINATOR]]
        )

    return point_estimate


def lin_reg_variance_delta(row, **kwargs):
    y = row[kwargs[NUMERATOR]]
    n = row[kwargs[DENOMINATOR]]

    XX = unlist(row[kwargs[FEATURE_SUMSQ]])
    X = unlist(row[kwargs[FEATURE]])
    Xy = unlist(row[kwargs[FEATURE_CROSS]])

    sample_var = XX / n - dfmatmul(X / n, X / n)
    sample_cov = Xy / n - dfmatmul(X / n, y / n)
    b = np.atleast_2d(row[REGRESSION_PARAM])
    variance2 = np.matmul(np.transpose(b), np.matmul(sample_var, b)).item()
    variance3 = -2 * np.matmul(np.transpose(b), sample_cov).item()

    return variance2 + variance3


def variance(df: DataFrame, **kwargs) -> Series:
    variance1 = z_test_computer.variance(df, **kwargs)

    if kwargs[FEATURE] in df:
        return variance1 + df.apply(lin_reg_variance_delta, axis=1, **kwargs)
    else:
        return variance1


def add_point_estimate_ci(df: DataFrame, **kwargs: Dict) -> DataFrame:
    return z_test_computer.add_point_estimate_ci(df, **kwargs)


def std_err(df: DataFrame, **kwargs: Dict) -> DataFrame:
    return z_test_computer.std_err(df, **kwargs)


def p_value(df: DataFrame, **kwargs: Dict) -> DataFrame:
    return z_test_computer.p_value(df, **kwargs)


def ci(df: DataFrame, alpha_column: str, **kwargs: Dict) -> DataFrame:
    return z_test_computer.ci(df, alpha_column, **kwargs)


def powered_effect(
    df: DataFrame,
    z_alpha: float,
    z_power: float,
    binary: bool,
    non_inferiority: bool,
    avg_column: float,
    var_column: float,
) -> Series:
    return z_test_computer.powered_effect(df, z_alpha, z_power, binary, non_inferiority, avg_column, var_column)


def required_sample_size(
    binary: Union[Series, bool],
    non_inferiority: Union[Series, bool],
    hypothetical_effect: Union[Series, float],
    control_avg: Union[Series, float],
    control_var: Union[Series, float],
    z_alpha: float = None,
    kappa: float = None,
    proportion_of_total: Union[Series, float] = None,
    z_power: float = None,
) -> Union[Series, float]:
    return z_test_computer.required_sample_size(
        binary,
        non_inferiority,
        hypothetical_effect,
        control_avg,
        control_var,
        z_alpha,
        kappa,
        proportion_of_total,
        z_power,
    )



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import numpy as np
from itertools import product


def example_data_binomial():
    """
    Returns an output dataframe with categorical
    features (country and test variation), and orginal features (date),
    as well as number of successes and total observations for each combination
    """
    countries = ["ca", "us"]
    dates = pd.date_range("2018-01-01", "2018-02-01")
    variation_names = ["test", "control", "test2"]

    # test ca, test us, control ca, control us, test2 ca, test2 us
    success_rates = [0.3, 0.32, 0.24, 0.22, 0.25, 0.42]
    n_observations = [50, 80, 30, 50, 40, 50]

    return_df = pd.DataFrame()

    for i, (country, variation) in enumerate(product(countries, variation_names)):
        df = pd.DataFrame({"date": dates})
        df["country"] = country
        df["variation_name"] = variation
        df["total"] = np.random.poisson(n_observations[i], size=len(dates))
        df["success"] = df["total"].apply(lambda x: np.random.binomial(x, success_rates[i]))
        return_df = pd.concat([return_df, df], axis=0)

    return return_df


def example_data_gaussian():
    df = pd.DataFrame(
        {
            "variation_name": [
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
                "test",
                "control",
                "test2",
            ],
            "nr_of_items": [
                500,
                8,
                100,
                510,
                8,
                100,
                520,
                9,
                104,
                530,
                7,
                100,
                530,
                8,
                103,
            ],
            "nr_of_items_sumsq": [
                2500,
                12,
                150,
                2510,
                13,
                140,
                2520,
                14,
                154,
                2530,
                15,
                160,
                2530,
                16,
                103,
            ],
            "users": [
                1010,
                22,
                150,
                1000,
                20,
                153,
                1030,
                23,
                154,
                1000,
                20,
                150,
                1040,
                21,
                155,
            ],
            "days_since_reg": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5],
        }
    )

    return df



# Copyright 2017-2020 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Union, Iterable

from pandas import DataFrame

from spotify_confidence.chartgrid import ChartGrid
from ..constants import NIM_TYPE


class ConfidenceGrapherABC(ABC):
    @abstractmethod
    def __init__(
        self,
        data_frame: DataFrame,
        numerator_column: str,
        denominator_column: str,
        categorical_group_columns: str,
        ordinal_group_column: str,
    ):
        pass

    @abstractmethod
    def plot_summary(self, summary_df: DataFrame, groupby: Union[str, Iterable]) -> ChartGrid:
        """Plot for each group in the data_frame:

        if ordinal level exists:
            line graph with area to represent confidence interval
        if categorical levels:
            Interval plots of confidence intervals by group

        Args:
            summary_df (DataFrame): A data frame produced by a
            ConfidenceComputer's summary method

        Returns:
            ChartGrid object.
        """
        pass

    @abstractmethod
    def plot_difference(
        self,
        difference_df: DataFrame,
        absolute: bool,
        groupby: Union[str, Iterable],
        nims: NIM_TYPE,
        use_adjusted_intervals: bool,
        split_plot_by_groups: bool,
    ) -> ChartGrid:
        """Plot representing the difference between group 1 and 2 with
        confidence intervals.

        Args:
            difference_df (DataFrame): A dataframe produced by a
            ConfidenceComputer's difference method

        Returns:
            Chartify Chart object.
            :param groupby:
        """

    @abstractmethod
    def plot_differences(
        self,
        difference_df: DataFrame,
        absolute: bool,
        groupby: Union[str, Iterable],
        nims: NIM_TYPE,
        use_adjusted_intervals: bool,
        split_plot_by_groups: bool,
    ) -> ChartGrid:
        """Plot representing the difference between group 1 and 2 with
        confidence intervals.

        Args:
            difference_df (DataFrame): A dataframe produced by a
            ConfidenceComputer's difference method

        Returns:
            Chartify Chart object.
            :param groupby:
        """

    @abstractmethod
    def plot_multiple_difference(
        self,
        difference_df: DataFrame,
        absolute: bool,
        groupby: Union[str, Iterable],
        level_as_reference: bool,
        nims: NIM_TYPE,
        use_adjusted_intervals: bool,
        split_plot_by_groups: bool,
    ) -> ChartGrid:
        """Compare level to all other groups or, if level_as_reference = True,
        all other groups to level.

        Args:
            difference_df (DataFrame): A dataframe produced by a
            ConfidenceComputer's multiple_difference method

        Returns:
            ChartGrid object.
        """



"""Tests for `confidence` categorical variables."""

import numpy as np
import pandas as pd
import pytest

import spotify_confidence
from spotify_confidence.analysis.constants import (
    POINT_ESTIMATE,
    VARIANCE,
    NUMERATOR,
    NUMERATOR_SUM_OF_SQUARES,
    DENOMINATOR,
    SFX1,
    SFX2,
    INCREASE_PREFFERED,
    DECREASE_PREFFERED,
    DIFFERENCE,
    P_VALUE,
    CI_UPPER,
    CI_LOWER,
)
from spotify_confidence.analysis.frequentist.confidence_computers import t_test_computer as computer


def chart_data(chart_object, series_name):
    """Retrieve data from chart object based on series name.

    Note: if there's only one series the name is None.

    Returns:
        Dictionary populated with data from the chart.
    """
    cannonical_series_name = chart_object.plot._cannonical_series_name(series_name)
    return chart_object.figure.select(cannonical_series_name)[0].data


class TestCategorical(object):
    def setup(self):
        np.random.seed(123)

        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test", "control", "test2"],
                "nr_of_items": [1969, 312, 2955, 195, 24, 330],
                "nr_of_items_sumsq": [5767, 984, 8771, 553, 80, 1010],
                "users": [1009, 104, 1502, 100, 10, 150],
                "country": [
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                ],
            }
        )

        self.test = spotify_confidence.StudentsTTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["country", "variation_name"],
            interval_size=0.95,
        )

    def test_sample_variance(self):
        n = 1000
        x = np.random.randn(n)
        var = x.var()

        sum = np.sum(x)
        sum_squared = np.sum(x * x)

        kwargs = {NUMERATOR: "sum", NUMERATOR_SUM_OF_SQUARES: "sumsq", DENOMINATOR: "n"}
        var_to_verify = pd.DataFrame(
            {"sum": [sum], "sumsq": [sum_squared], POINT_ESTIMATE: [x.mean()], "n": [n]}
        ).apply(lambda row: computer.variance(row, **kwargs), axis=1)

        assert np.allclose(var_to_verify, var, rtol=2e-3)

    def test_standard_error_diff_mean_from_sums(self):
        n1 = int(1e6)
        mu1 = 1
        sigma1 = 1
        x1 = np.random.normal(mu1, sigma1, n1)

        n2 = int(1e6)
        mu2 = 2
        sigma2 = 2
        x2 = np.random.normal(mu2, sigma2, n2)

        std_diff = np.sqrt(sigma1**2 / n1 + sigma2**2 / n2)

        kwargs = {DENOMINATOR: "n"}
        diff_se = computer.std_err(
            pd.DataFrame(
                {
                    VARIANCE + SFX1: [x1.var()],
                    VARIANCE + SFX2: [x2.var()],
                    "n" + SFX1: [n1],
                    "n" + SFX2: [n2],
                    "method": ["t-test"],
                }
            ),
            **kwargs,
        )

        assert np.allclose(std_diff, diff_se, atol=1e-5)

    def test_summary(self):
        summary = self.test.summary()

        assert np.array_equal(summary.country, np.array(["us", "us", "us", "gb", "gb", "gb"]))
        assert np.array_equal(summary.point_estimate, self.data.nr_of_items / self.data.users)

        assert np.allclose(
            summary["ci_lower"],
            np.array([1.86607468, 2.86724041, 1.89633269, 1.68789181, 1.27143773, 1.97725383]),
            rtol=1e-3,
        )

        assert np.allclose(
            summary["ci_upper"],
            np.array([2.03679946, 3.13275959, 2.03842097, 2.21210819, 3.52856227, 2.42274617]),
            rtol=2e-3,
        )

    def test_summary_plot(self):
        """Area plot tests"""
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1
        ch = self.test.summary_plot(groupby="country")
        assert len(ch.charts) == 2

    def test_p_value_is_symmetric_and_correct(self):
        df = pd.DataFrame(
            {
                "group_name": ["Control", "Test"],
                "users": [48351, 50571],
                "nr_of_items": [1.438602e06, 1.521974e06],
                "nr_of_items_sumsq": [7.330581e07, 7.862121e07],
            }
        )

        ttest = spotify_confidence.StudentsTTest(
            data_frame=df,
            denominator_column="users",
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            categorical_group_columns="group_name",
        )

        diff_summary_1 = ttest.difference(level_1="Control", level_2="Test")
        diff_summary_2 = ttest.difference(level_1="Test", level_2="Control")

        assert diff_summary_1["p-value"].iloc[0] == diff_summary_2["p-value"].iloc[0]
        assert np.isclose(diff_summary_1["p-value"].iloc[0], 0.03334, 0.01)

    def test_difference(self):
        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], -1.0485629, rtol=1e-3)
        assert np.allclose(diff["ci_lower"].iloc[0], -1.205402, rtol=1e-3)
        assert np.allclose(diff["ci_upper"].iloc[0], -0.891723, rtol=1e-3)
        assert np.allclose(diff["p-value"].iloc[0], 0.00000, rtol=1e-3)

        diff = self.test.difference(("gb", "control"), ("gb", "test2"))
        assert np.allclose(diff["difference"].iloc[0], -0.2, rtol=1e-3)
        assert np.allclose(diff["ci_lower"].iloc[0], -1.340533, rtol=1e-3)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.940533, rtol=1e-3)
        assert np.allclose(diff["p-value"].iloc[0], 0.704021, rtol=1e-3)

        diff = self.test.difference(("us", "control"), ("us", "test"), absolute=False)
        assert np.allclose(diff["difference"].iloc[0], -0.349520, rtol=1e-3)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.4018008, rtol=1e-3)
        assert np.allclose(diff["ci_upper"].iloc[0], -0.297241, rtol=1e-3)
        assert np.allclose(diff["p-value"].iloc[0], 0.00000, rtol=1e-3)

        diff = self.test.difference("control", "test", groupby="country")
        assert np.allclose(diff["difference"], np.array([-0.450000, -1.048563]), rtol=1e-3)
        assert np.allclose(diff["ci_lower"], np.array([-1.595350, -1.205963]), rtol=1e-3)
        assert np.allclose(diff["ci_upper"], np.array([0.695350, -0.891163]), rtol=1e-3)
        assert np.allclose(diff["p-value"], np.array([0.4030961, 0.000000]), rtol=1e-3)

    def test_difference_with_interval_sizes(self):
        """
        https: // www.quantitativeskills.com / sisa / statistics / t - test.htm
        was used to validate results
        """
        self.test._confidence_computer._interval_size = 0.99

        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], -1.0485629, rtol=1e-3)
        assert np.allclose(diff["ci_lower"].iloc[0], -1.255391, rtol=1e-3)
        assert np.allclose(diff["ci_upper"].iloc[0], -0.841735, rtol=1e-3)
        assert np.allclose(diff["p-value"].iloc[0], 0.00000, rtol=1e-3)

        self.test._confidence_computer._interval_size = 0.999
        diff = self.test.difference(("us", "control"), ("us", "test"))
        assert np.allclose(diff["difference"].iloc[0], -1.0485629, rtol=1e-3)
        assert np.allclose(diff["ci_lower"].iloc[0], -1.314140, rtol=1e-3)
        assert np.allclose(diff["ci_upper"].iloc[0], -0.782024, rtol=1e-3)
        assert np.allclose(diff["p-value"].iloc[0], 0.00000, rtol=1e-3)

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot(("bad_value", "bad_value"), ("bad_value", "bad_value"))

        ch = self.test.difference_plot(("us", "control"), ("us", "test"))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot("control", "test", groupby="country")
        assert len(ch.charts) == 1

    def test_multiple_difference(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference(("bad_value", "bad_value"), level_as_reference=False)

        diff = self.test.multiple_difference(("us", "control"), level_as_reference=False)
        assert len(diff) > 0

        diff = self.test.multiple_difference("test", groupby="country", level_as_reference=False)
        assert len(diff) > 0

    def test_multiple_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_plot(
                ("bad_value", "bad_value"), ("bad_value", "bad_value"), level_as_reference=False
            )

        ch = self.test.multiple_difference_plot(("us", "control"), level_as_reference=False)
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot("test", groupby="country", level_as_reference=False)
        assert len(ch.charts) == 1

    def test_achieved_power(self):
        power_df = self.test.achieved_power(level_1=("us", "test2"), level_2=("us", "test"), mde=0.2, alpha=0.05)
        power = power_df.achieved_power
        expected_power = 0.9409124  # Based on G*Power calculation

        assert np.allclose(power, expected_power, 0.005)


class TestOrdinal(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "nr_of_items": [500, 8, 100, 510, 8, 100, 520, 9, 104, 530, 7, 100, 530, 8, 103],
                "nr_of_items_sumsq": [2500, 12, 150, 2510, 13, 140, 2520, 14, 154, 2530, 15, 160, 2530, 16, 103],
                "users": [1010, 22, 150, 1000, 20, 153, 1030, 23, 154, 1000, 20, 150, 1040, 21, 155],
                "days_since_reg": [1, 1, 1, 2, 2, 2, 3, 3, 3, 4, 4, 4, 5, 5, 5],
            }
        )

        self.test = spotify_confidence.StudentsTTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns="variation_name",
            ordinal_group_column="days_since_reg",
            interval_size=0.95,
        )

    def test_summary(self):
        summary = self.test.summary()

        assert np.array_equal(summary.point_estimate, self.data.nr_of_items / self.data.users)

        assert np.allclose(
            summary["ci_lower"],
            np.array(
                [
                    0.40279383,
                    0.07191692,
                    0.54600746,
                    0.41687347,
                    0.06387915,
                    0.54166658,
                    0.41429269,
                    0.09286814,
                    0.5575301,
                    0.43689002,
                    -0.03036827,
                    0.53897297,
                    0.4198777,
                    0.01463557,
                    0.58935338,
                ]
            ),
            rtol=1e-3,
        )

        assert np.allclose(
            summary["ci_upper"],
            np.array(
                [
                    0.58730518,
                    0.6553558,
                    0.78732587,
                    0.60312653,
                    0.73612085,
                    0.76552296,
                    0.59541605,
                    0.68974056,
                    0.79311926,
                    0.62310998,
                    0.73036827,
                    0.79436036,
                    0.59935306,
                    0.7472692,
                    0.73967887,
                ]
            ),
            rtol=1e-3,
        )

    def test_summary_plot(self):
        """Area plot tests"""
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1

        ch = self.test.summary_plot().charts[0]
        np.array_equal(chart_data(ch, "control")["days_since_reg"], np.array([1, 2, 3, 4, 5, 5, 4, 3, 2, 1]))

    def test_difference(self):
        with pytest.raises(ValueError):
            self.test.difference(("control", "us"), ("test", "usf"))

        diff = self.test.difference(("control", 1), ("test", 1))
        assert np.allclose(diff["difference"].iloc[0], 0.131413, rtol=1e-3)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.172704, rtol=1e-3)
        assert np.allclose(diff["ci_upper"].iloc[0], 0.435531, rtol=1e-3)
        assert np.allclose(diff["p-value"].iloc[0], 0.382552, rtol=1e-3)

        diff = self.test.difference(("control", 1), ("test", 1), absolute=False)
        assert np.allclose(diff["difference"].iloc[0], 0.361386, rtol=1e-3)
        assert np.allclose(diff["ci_lower"].iloc[0], -0.474938, rtol=1e-3)
        assert np.allclose(diff["ci_upper"].iloc[0], 1.19771, rtol=1e-3)
        assert np.allclose(diff["p-value"].iloc[0], 0.382552, rtol=1e-3)

        diff = self.test.difference("control", "test", groupby="days_since_reg")
        assert np.allclose(diff["difference"], np.array([0.13141314, 0.11, 0.11355002, 0.18, 0.128663]), rtol=1e-3)

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot(("control", 10), ("test", 10))

        ch = self.test.difference_plot(("control", 1), ("test", 1))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot("control", "test", groupby="days_since_reg")
        assert len(ch.charts) == 1

    def test_multiple_difference(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference(
                ("bad_value", "bad_value"), ("bad_value", "bad_value"), level_as_reference=False
            )

        diff = self.test.multiple_difference(("control", 1), level_as_reference=False)
        assert len(diff) > 0

        diff = self.test.multiple_difference("test", groupby="days_since_reg", level_as_reference=False)
        assert len(diff) > 0

    def test_multiple_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_plot(
                ("bad_value", "bad_value"), ("bad_value", "bad_value"), level_as_reference=False
            )

        ch = self.test.multiple_difference_plot(("control", 1), level_as_reference=False)
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot("test", groupby="days_since_reg", level_as_reference=False)
        assert len(ch.charts) == 1

    def test_achieved_power(self):
        power_df = self.test.achieved_power(level_1=("control", 1), level_2=("test", 1), mde=1, alpha=0.05)
        power = power_df.achieved_power
        expected_power = 0.8790801  # Based on G*Power calculation
        assert np.allclose(power, expected_power, 0.005)


class TestOrdinalPlusTwoCategorical(object):
    def setup(self):
        self.data = pd.DataFrame(
            {
                "variation_name": [
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                    "test",
                    "control",
                    "test2",
                ],
                "nr_of_items": [
                    500,
                    8,
                    100,
                    510,
                    8,
                    100,
                    520,
                    9,
                    104,
                    530,
                    7,
                    100,
                    530,
                    8,
                    103,
                    500,
                    8,
                    100,
                    510,
                    8,
                    100,
                    520,
                    9,
                    104,
                    530,
                    7,
                    100,
                    530,
                    8,
                    103,
                ],
                "nr_of_items_sumsq": [
                    1010,
                    32,
                    250,
                    1000,
                    30,
                    253,
                    1030,
                    33,
                    254,
                    1000,
                    30,
                    250,
                    1040,
                    31,
                    255,
                    1010,
                    22,
                    150,
                    1000,
                    20,
                    153,
                    1030,
                    23,
                    154,
                    1000,
                    20,
                    150,
                    1040,
                    21,
                    155,
                ],
                "users": [
                    2010,
                    42,
                    250,
                    2000,
                    40,
                    253,
                    2030,
                    43,
                    254,
                    2000,
                    40,
                    250,
                    2040,
                    41,
                    255,
                    1010,
                    22,
                    150,
                    1000,
                    20,
                    153,
                    1030,
                    23,
                    154,
                    1000,
                    20,
                    150,
                    1040,
                    21,
                    155,
                ],
                "days_since_reg": [
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                    1,
                    1,
                    1,
                    2,
                    2,
                    2,
                    3,
                    3,
                    3,
                    4,
                    4,
                    4,
                    5,
                    5,
                    5,
                ],
                "country": [
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                    "gb",
                ],
            }
        )

        self.test = spotify_confidence.StudentsTTest(
            self.data,
            numerator_column="nr_of_items",
            numerator_sum_squares_column="nr_of_items_sumsq",
            denominator_column="users",
            categorical_group_columns=["variation_name", "country"],
            ordinal_group_column="days_since_reg",
        )

    def test_summary_plot(self):
        ch = self.test.summary_plot()
        assert len(ch.charts) == 1

        ch = self.test.summary_plot(groupby=["country"])
        assert len(ch.charts) == 2

        ch = self.test.summary_plot(groupby=["days_since_reg"])
        assert len(ch.charts) == 5

        ch = self.test.summary_plot(groupby=["country", "days_since_reg"])
        assert len(ch.charts) == 10

    def test_difference(self):
        with pytest.raises(ValueError):
            self.test.difference(("control", "us"), ("test", "usf"), groupby="days_since_reg")

        df = self.test.difference(level_1=("test", "us"), level_2=("control", "us"), groupby="days_since_reg")
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(level_1=("test", "us"), level_2=("control", "us"), groupby=["days_since_reg"])
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(level_1=("test", 1), level_2=("control", 1), groupby=["country"])
        assert len(df) == 2
        assert "country" in df.columns

    def test_differece_with_nims(self):
        df = self.test.difference(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby="days_since_reg",
            non_inferiority_margins=(0.01, INCREASE_PREFFERED),
        )
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby=["days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(df) == 5
        assert "days_since_reg" in df.columns

        df = self.test.difference(
            level_1=("test", 1),
            level_2=("control", 1),
            groupby=["country"],
            non_inferiority_margins={"us": (0.01, INCREASE_PREFFERED), "gb": (0.05, INCREASE_PREFFERED)},
        )
        assert len(df) == 2
        assert "country" in df.columns

        df = self.test.difference(
            level_1="test",
            level_2="control",
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(df) == 10
        assert "country" in df.columns
        assert "days_since_reg" in df.columns

        nims = {
            ("us", 1): (0.01, DECREASE_PREFFERED),
            ("us", 2): (0.1, INCREASE_PREFFERED),
            ("us", 3): (0.2, DECREASE_PREFFERED),
            ("us", 4): (0.5, INCREASE_PREFFERED),
            ("us", 5): (0.99, DECREASE_PREFFERED),
            ("gb", 1): (1.01, INCREASE_PREFFERED),
            ("gb", 2): (2.01, DECREASE_PREFFERED),
            ("gb", 3): (3.01, INCREASE_PREFFERED),
            ("gb", 4): (4.01, DECREASE_PREFFERED),
            ("gb", 5): (5.01, INCREASE_PREFFERED),
        }
        df = self.test.difference(
            level_1="test", level_2="control", groupby=["country", "days_since_reg"], non_inferiority_margins=nims
        )
        assert len(df) == 10
        assert "country" in df.columns
        assert "days_since_reg" in df.columns

        df = self.test.multiple_difference(
            level="control",
            level_as_reference=True,
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=nims,
        )
        assert len(df) == 20

    def test_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.difference_plot(("control", "us", 10), ("test", "us", 10))

        ch = self.test.difference_plot(level_1=("control", "us", 1), level_2=("test", "us", 1))
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1=("control", "us"), level_2=("test", "us"), groupby="days_since_reg")
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1="control", level_2="test", groupby=["country", "days_since_reg"])
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1=("control", 1), level_2=("test", 1), groupby="country")
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1=("control", "us"), level_2=("test", "us"), groupby="days_since_reg")
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(level_1="control", level_2="test", groupby=["country", "days_since_reg"])
        assert len(ch.charts) == 1

    def test_multiple_difference_plot(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference_plot(level="control")

        ch = self.test.multiple_difference_plot(
            level="control", groupby=["country", "days_since_reg"], level_as_reference=True
        )
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot(
            level="us", groupby=["variation_name", "days_since_reg"], level_as_reference=True
        )
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot(
            level=1, groupby=["country", "variation_name"], level_as_reference=True
        )
        assert len(ch.charts) == 1

    def test_differece_plot_with_nims(self):
        ch = self.test.difference_plot(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby="days_since_reg",
            non_inferiority_margins=(0.01, INCREASE_PREFFERED),
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1=("test", "us"),
            level_2=("control", "us"),
            groupby=["days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1=("test", 1),
            level_2=("control", 1),
            groupby=["country"],
            non_inferiority_margins={"us": (0.01, INCREASE_PREFFERED), "gb": (0.05, INCREASE_PREFFERED)},
        )
        assert len(ch.charts) == 1

        ch = self.test.difference_plot(
            level_1="test",
            level_2="control",
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=(0.01, DECREASE_PREFFERED),
        )
        assert len(ch.charts) == 1

        nims = {
            ("us", 1): (0.01, DECREASE_PREFFERED),
            ("us", 2): (0.1, INCREASE_PREFFERED),
            ("us", 3): (0.2, DECREASE_PREFFERED),
            ("us", 4): (0.5, INCREASE_PREFFERED),
            ("us", 5): (0.99, DECREASE_PREFFERED),
            ("gb", 1): (1.01, INCREASE_PREFFERED),
            ("gb", 2): (2.01, DECREASE_PREFFERED),
            ("gb", 3): (3.01, INCREASE_PREFFERED),
            ("gb", 4): (4.01, DECREASE_PREFFERED),
            ("gb", 5): (5.01, INCREASE_PREFFERED),
        }
        ch = self.test.difference_plot(
            level_1="test", level_2="control", groupby=["country", "days_since_reg"], non_inferiority_margins=nims
        )
        assert len(ch.charts) == 1

        ch = self.test.multiple_difference_plot(
            level="control",
            level_as_reference=True,
            groupby=["country", "days_since_reg"],
            non_inferiority_margins=nims,
        )
        assert len(ch.charts) == 1


class TestCategoricalBinomialData(object):
    def setup(self):
        np.random.seed(123)

        self.data = pd.DataFrame(
            {
                "variation_name": ["test", "control", "test2", "test", "control", "test2"],
                "success": [500, 42, 1005, 50, 4, 100],
                "total": [1009, 104, 1502, 100, 10, 150],
                "country": [
                    "us",
                    "us",
                    "us",
                    "gb",
                    "gb",
                    "gb",
                ],
            }
        )

        self.test = spotify_confidence.StudentsTTest(
            self.data,
            numerator_column="success",
            numerator_sum_squares_column="success",
            denominator_column="total",
            categorical_group_columns=["country", "variation_name"],
        )

    def test_summary(self):
        """Area plot tests"""

        summary = self.test.summary()
        assert np.array_equal(summary.country, np.array(["us", "us", "us", "gb", "gb", "gb"]))
        assert np.array_equal(summary.point_estimate, self.data.success / self.data.total)
        assert np.allclose(
            summary["ci_lower"], np.array([0.464653, 0.308424, 0.645293, 0.400789, 0.049548, 0.590610]), rtol=1e-3
        )
        assert np.allclose(
            summary["ci_upper"], np.array([0.526427, 0.499269, 0.692923, 0.599211, 0.750452, 0.742723]), rtol=1e-3
        )

    def test_multiple_difference(self):
        with pytest.raises(ValueError):
            self.test.multiple_difference(("bad_value", "bad_value"), level_as_reference=False)

        diff = self.test.multiple_difference(("us", "control"), level_as_reference=False)
        assert np.allclose(
            diff["adjusted p-value"],
            np.array([1e00, 8.36850129e-01, 1.49172324e-04, 3.62369184e-01, 2.27154393e-06]),
            rtol=1e-3,
        )
        assert np.allclose(
            diff["p-value"],
            np.array([9.81516149e-01, 1.67370026e-01, 2.98344648e-05, 7.24738369e-02, 4.54308787e-07]),
            rtol=1e-3,
        )
        assert np.allclose(
            diff["adjusted ci_lower"], np.array([-0.501747, -0.276600, -0.422946, -0.224093, -0.395207]), rtol=1e-3
        )
        assert np.allclose(
            diff["adjusted ci_upper"], np.array([0.509440, 0.084292, -0.102695, 0.040705, -0.135317]), rtol=1e-3
        )

        diff = self.test.multiple_difference("test", groupby="country", level_as_reference=False)
        assert np.allclose(diff["adjusted p-value"], np.array([1, 0.035594, 0.289895, 0.0]), rtol=1e-3)
        assert np.allclose(diff["p-value"], np.array([0.55155672, 0.00889848, 0.07247384, 0.0]), rtol=1e-3)
        assert np.allclose(
            diff["adjusted ci_lower"], np.array([-0.385571, -0.325682, -0.036587, -0.223262]), rtol=1e-3
        )
        assert np.allclose(diff["adjusted ci_upper"], np.array([0.585571, -0.0076513, 0.219975, -0.123874]), rtol=1e-3)


class TestWithNims(object):
    def setup(self):
        self.data = pd.DataFrame(
            [
                {
                    "group": "1",
                    "count": 5000,
                    "sum": 10021.0,
                    "sum_of_squares": 25142.0,
                    "avg": 2.004210,
                    "var": 1.0116668,
                },
                {
                    "group": "2",
                    "count": 5000,
                    "sum": 9892.0,
                    "sum_of_squares": 24510.0,
                    "avg": 1.978424,
                    "var": 0.9881132,
                },
            ]
        )

        self.test = spotify_confidence.StudentsTTest(
            self.data,
            numerator_column="sum",
            numerator_sum_squares_column="sum_of_squares",
            denominator_column="count",
            categorical_group_columns="group",
            interval_size=0.99,
        )

    def test_compare_series_non_inferiority_improve_postitive(self):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(0.02, "increase"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert diff[CI_UPPER].values[0] == float("inf")
        np.testing.assert_almost_equal(diff[CI_LOWER].values[0], -0.0723, 3)
        assert diff[P_VALUE].values[0] > 0.01

    def test_compare_series_non_inferiority_improve_negative(self):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(0.02, "decrease"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert diff[CI_LOWER].values[0] == -float("inf")
        np.testing.assert_almost_equal(diff[CI_UPPER].values[0], 0.0207, 3)
        assert diff[P_VALUE].values[0] < 0.01

    def test_one_sided_ztest_positive(self):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(None, "increase"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert diff[CI_UPPER].values[0] == float("inf")
        np.testing.assert_almost_equal(diff[CI_LOWER].values[0], -0.0723, 3)
        assert diff[P_VALUE].values[0] > 0.01

    def test_one_sided_ztest_negative(self):
        summary = self.test.summary()
        control_avg = self.data.query("group == '1'").avg.values[0]
        assert np.allclose(control_avg, summary.query("group == '1'")[POINT_ESTIMATE])

        diff = self.test.difference(level_1="1", level_2="2", non_inferiority_margins=(None, "decrease"))

        np.testing.assert_almost_equal(diff[DIFFERENCE].values[0], -0.0258, 3)
        assert diff[CI_LOWER].values[0] == -float("inf")
        np.testing.assert_almost_equal(diff[CI_UPPER].values[0], 0.0207, 3)
        assert diff[P_VALUE].values[0] > 0.01
